#ifndef __NeuralNetwork_H
#define __NeuralNetwork_H

#include <stdio.h>
#include "idx.h"

#define _CRT_SECURE_NO_WARNINGS 

#define SCALEDTANH(X)  (1.7159047*tanh(0.66666667*(X)))
#define DSCALEDTANH(S) (0.66666667/1.7159047*(1.7159047+(S))*(1.7159047-(S)))

enum LAYER_TYPE {input=0, fully_connected, convolutional, max_pooling, one_kernel_convolutional, avg_pooling, vertical_pooling=6};
enum ACTIVATION_TYPE {scaled_tanh, soft_max};
enum CONNECTION_TABLE {full,not_full}; // use more conncection table types later on

typedef struct{
	int	nNeurons;
	int nMaps;
	int	nWeights;
	int Kx,Ky,Mx,My,Sx,Sy;
	int *C;						//connection table
	int *BP_C;					//connection table for BP
	double			*W;
	LAYER_TYPE		type;
	ACTIVATION_TYPE	activation;
	CONNECTION_TABLE C_type;
	double			*y;
	double			*d;
	int					nMC;		//number of map connections
	int					*max_pos;		//contains the index of neuron with max value from the previous layer	(for max_pooling)
	int					IP;				//number of new maps (obtained by image processing) for each input map. VALID ONLY FOR INPUT LAYER!
	double			*IPkernels;	//kernel values
	int					IPk;				//size of the kernel
}Layer;

class NeuralNetwork{
private:
	int version;
	double eta;
	double etaStart;
	double etaStop;
	double etaFactor;
	float *partial;
	unsigned  count;
	int		epoch;
	double	best_test,best_validation;
	char	best_test_name[1000],best_validation_name[1000];
public:
	Layer *listLayers;
	int		nLayers;
	char	NNname[1000];
	double minERR, maxERR, ERR;		//CE or MSE
	NeuralNetwork();

	void TrainCPU(IDXHandler idxh_train, IDXHandler idxh_validation, IDXHandler idxh_test);
	void TrainOneEpochCPU(IDXHandler idxh);
	double determineErrorRateCPU(IDXHandler idxh, Error *errorvector);
	double determineErrorRateCPU_BIG(IDXHandler idxh, Error *errorvector);

	int  getOutputSize(void);
	int  getInputSize(void);

	void setEta(double etaStart, double etaStop, double etaFactor);
	void setLoad(bool Load);
	void create(const char* filename);
	
private:
	void createLayer(int idxLayer, FILE *fin);
	void createInputLayer(FILE *fin);
	void createFullyConnectedLayer(int idxLayer, FILE *fin);
	void createConvolutionalLayer(int idxLayer, FILE *fin);
	void createPoolingLayer(int idxLayer, FILE *fin);
	
	void correctWeightsCPU(double *image, unsigned char label);

	void setInput(unsigned char *image, unsigned char width, unsigned char height);
	void FPLayerCPU(int index);
	void FPFullyConnectedLayerCPU(int index);
	void FPConvolutionalLayerCPU(int index);
	void FPMaxPoolingLayerCPU(int index);
	
	int getClass(double *v);
	
	double *getDesiredOutput(unsigned char label);
	void BPLayerCPU(int idxlay);
	void BPFullyConnectedLayerCPU(int idxlay);
	void BPConvolutionalLayerCPU(int idxlayer);
	void BPMaxPoolingLayerCPU(int idxlayer);
	
	void delete_files(char *base, char *type);
	void write_error_files(char *type, char *base, Error *errorvector, int errors, IDXHandler idxh);
	void updateBestResults(int write_test, int write_validation, Error *errorvectorV, Error *errorvectorT, float validation_err, float test_err, IDXHandler &idxh_validation, IDXHandler &idxh_test, bool isGPU);

};

#endif
