#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string>
#include "idx.h"
#include "neuralnetwork.h"
using namespace std;

void run_trainnet();

int main(int argc, char** argv){
	printf("Version: 0.50\n");	
	run_trainnet();
	return 0;
}

void run_trainnet(){
	time_t t; srand((unsigned) time(&t));
	
	//initialize network
	NeuralNetwork nn = NeuralNetwork();

	//load and set eta
	double etaStart=0.001;
	double etaStop=0.000001;
	double etaFactor=0.9;

	//nn.create("29x29-20C4.0-MP2-40C5.0-MP3-150N.txt");
	nn.create("29x29-5C4.0-MP2-10C5.0-MP3-50N.txt");

	nn.setEta(etaStart,etaStop,etaFactor);

	IDXHandler idxh_train(100);
	IDXHandler idxh_validation(1000);
	IDXHandler idxh_test(1000);
	
	idxh_train.nImages			=	0;
	idxh_validation.nImages		=	0;
	idxh_test.nImages			=	0;
	
	idxh_train.loadIDXimages("29x29-train-images.idx");
	idxh_train.loadIDXlabels("29x29-train-labels.idx");
	idxh_train.allocateBuffer();
	idxh_validation.loadIDXimages("29x29-train-images.idx");
	idxh_validation.loadIDXlabels("29x29-train-labels.idx");
	idxh_validation.allocateBuffer();
	idxh_test.loadIDXimages("29x29-test-images.idx");
	idxh_test.loadIDXlabels("29x29-test-labels.idx");
	idxh_test.allocateBuffer();
	
	if(idxh_train.n_labels!=idxh_test.n_labels || idxh_train.n_labels!=idxh_validation.n_labels || idxh_validation.n_labels!=idxh_test.n_labels){
		printf("The number of classes in training/validation/test label files is different!\n");
		printf("Train: %d\tValidation: %d\tTest: %d",idxh_train.n_labels, idxh_validation.n_labels, idxh_test.n_labels);
		exit(1);
	}
	
	if(idxh_train.n_labels!=nn.getOutputSize()){
		printf("The number of classes (%d) is different than the number of output neurons (%d)!",idxh_train.n_labels,nn.getOutputSize());
		exit(1);
	}
	
	double elapsedTime;
	clock_t t1,t2;
	t1 = clock();

	nn.TrainCPU(idxh_train,idxh_validation,idxh_test);
	
	t2 = clock();
	
	elapsedTime = (double)(t2-t1)/CLOCKS_PER_SEC;

	idxh_validation.unloadIDX();
	idxh_train.unloadIDX();
	idxh_test.unloadIDX();
}
