#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include <float.h>
#include <set>
#include "idx.h"
#include "neuralnetwork.h"
using namespace std;

#define _CRT_SECURE_NO_WARNINGS 

NeuralNetwork::NeuralNetwork(){
	eta = 0.0;
	nLayers = 0;
	listLayers = NULL;
	minERR = -1.0;
	maxERR = 1e+10;
	ERR = 0.0;
	best_test_name[0]=0; 
	best_validation_name[0]=0;
}

void NeuralNetwork::create(const char* filename){
	FILE *fin=fopen(filename,"rt");
	if(fin==NULL){fprintf(stderr,"Error @create() : cannot open '%s' file.\7\n",filename);exit(1);}
	fscanf(fin,"Name: %s\n\n",NNname);

	fscanf(fin,"Number of layers: %d\n",&nLayers);
	listLayers=(Layer*)malloc(nLayers * sizeof(Layer));
	for(int i=0;i<nLayers;i++)
		createLayer(i,fin);
	fclose(fin);
}

void NeuralNetwork::createLayer(int idxLayer, FILE *fin){
	char s[1001];
	char layer_types[7][50]={"input", "fully_connected", "convolutional", "one_kernel_convolutional", "max_pooling", "avg_pooling","vertical_pooling"};
	int lt=-1;
	fscanf(fin,"Layer type: %s\n",s);
	for(int i=0;i<7;i++)
		if(strcmp(s,layer_types[i])==0){lt=i;break;}
	switch(lt){
		case 0:	createInputLayer(fin); break;
		case 1:	createFullyConnectedLayer(idxLayer,fin); break;
		case 2:	createConvolutionalLayer(idxLayer,fin); break;
		case 4:	listLayers[idxLayer].type=max_pooling; createPoolingLayer(idxLayer,fin); break;
		default: printf("Unknown layer type (%s) in %s!\7\n",s,__FUNCTION__);exit(1);break;
	}
}

void NeuralNetwork::createInputLayer(FILE *fin){
	Layer *layer=&listLayers[0];
	layer->type=input;
	fscanf(fin,"Number of neurons: %d\n",&(layer->nNeurons));
	fscanf(fin,"Number of maps: %d\n",&layer->nMaps);
	fscanf(fin,"Size: %d %d\n",&layer->Mx,&layer->My);
	if(layer->nNeurons != layer->nMaps * layer->Mx * layer->My){printf("Error @createInputLayer() : # neurons is wrong! Check if N==Mx*My.\n\7");exit(1);}
	
	if(fscanf(fin,"Image processing: %d\n",&layer->IP)!=1){printf("Error @createInputLayer() : Image processing is missing!\n\7");exit(1);}
	if(layer->IP==0)return;
	fscanf(fin,"Kernel size: %d\n",&layer->IPk);
	if(layer->IPk%2!=1){printf("Kernel size %d has to be odd in createInputLayer()!\7\n",layer->IPk);exit(1);}
	layer->IPkernels = (double*)malloc(layer->IP*layer->IPk*layer->IPk*sizeof(double));
	if(layer->IPkernels==NULL){printf("Error allocating memory for IPkernels in createInputLayer()!\7\n");exit(1);}
	fscanf(fin,"Kernel values:\n");
	for(int i=0;i<layer->IP;i++)
		for(int j=0;j<layer->IPk*layer->IPk;j++)
			if(fscanf(fin,"%lf ",&(layer->IPkernels[i*(layer->IPk*layer->IPk)+j]))!=1){printf("Error reading kernels!\n\7");exit(1);};

	layer->nNeurons	*= layer->IP;	//updating #neurons to reflect IP maps
	layer->nMaps	*= layer->IP;

	//allocating space for states
	layer->y=(double*)malloc(layer->nNeurons * sizeof(double));
	if(layer->y==NULL){fprintf(stderr,"Error @createInputLayer() : memory allocation (y) for input layer.\n");return;}
	//allocating space for deltas
	layer->d=(double*)malloc(layer->nNeurons * sizeof(double));
	if(layer->d==NULL){fprintf(stderr,"Error @createInputLayer() : memory allocation (d) for input layer.\n");return;}
}

//fully connected layer
void NeuralNetwork::createFullyConnectedLayer(int idxLayer, FILE *fin){
	Layer *layer=&listLayers[idxLayer];
	layer->type=fully_connected;
	char s[1001];
	fscanf(fin,"Number of neurons: %d\n",&(layer->nNeurons));
	fscanf(fin,"Activation function: %s\n\n",s);
	if(strcmp(s,"scaled_tanh")==0)
		layer->activation=scaled_tanh;
	else
		if(strcmp(s,"soft_max")==0)
			layer->activation=soft_max;
		else;//error
	//create W
	layer->nWeights=layer->nNeurons * (listLayers[idxLayer-1].nNeurons+1);
	layer->W=(double*)malloc(layer->nWeights * sizeof(double));
	if(layer->W==NULL){fprintf(stderr,"Error @createFullyConnectedLayer() : memory allocation (W) for layer %d.\n",idxLayer);return;}
	
	//weight initialization
	for (int i=0; i < layer->nWeights; i++){
		layer->W[i]=0.05*UNIFORM_PLUS_MINUS_ONE;
	}
#undef UNIFORM_PLUS_MINUS_ONE
	//allocating space for states
	layer->y=(double*)malloc(layer->nNeurons * sizeof(double));
	if(layer->y==NULL){fprintf(stderr,"Error @createFullyConnectedLayer() : memory allocation (y) for layer %d.\n",idxLayer);return;}
	//allocating space for deltas
	layer->d=(double*)malloc(layer->nNeurons * sizeof(double));
	if(layer->d==NULL){fprintf(stderr,"Error @createFullyConnectedLayer() : memory allocation (d) for layer %d.\n",idxLayer);return;}
	//other parameters to make it compatible with convolutional layers
	layer->nMaps=1;
	layer->Sx=layer->Sy=0;
}

void 
NeuralNetwork::createConvolutionalLayer(int idxLayer, FILE *fin){
	Layer *layer				= &listLayers[idxLayer];
	Layer *prev_layer	= &listLayers[idxLayer-1];
	int prev_nMaps		= prev_layer->nMaps;
	layer->type				= convolutional;
	char s[1001];

	fscanf(fin,"Number of maps: %d\n",&layer->nMaps);
	fscanf(fin,"Kernel size: %d %d\n",&layer->Kx,&layer->Ky);
	fscanf(fin,"Subsample size: %d %d\n",&layer->Sx, &layer->Sy);

	//checking
	if( (prev_layer->Mx - layer->Kx) % (layer->Sx+1)!=0 ||
		(prev_layer->My - layer->Ky) % (layer->Sy+1)!=0 ){
			fprintf(stderr,"Convolution out of bound on layer %d! Check kernel (%d-%d), input (%d-%d) and subsampling (%d-%d) sizes!", idxLayer, layer->Kx, layer->Ky, prev_layer->Mx, prev_layer->My, layer->Sx, layer->Sy);
			exit(1);
	}
	//computing the current map sizes
	layer->Mx = (prev_layer->Mx - layer->Kx) / (layer->Sx+1) + 1;
	layer->My = (prev_layer->My - layer->Ky) / (layer->Sy+1) + 1;

	//allocating connection table
	layer->C=(int*)malloc(layer->nMaps*(prev_nMaps+1)*sizeof(int));
	if(layer->C==NULL){fprintf(stderr,"Error @createConvolutionalLayer() : memory allocation (C) for layer %d.\n",idxLayer);return;}
	fscanf(fin,"Connection table: %s\n",s);
	// now either read the connection table if the string s is 'not_full' otherwise create a full conection table
	if (strcmp(s,"not_full")==0){
		layer->C_type=not_full;
		for(int i=0;i<layer->nMaps;i++){
			fscanf(fin,"%d",&layer->C[i*(prev_nMaps+1)+0]);
			if(layer->C[i*(prev_nMaps+1)+0]<0 || layer->C[i*(prev_nMaps+1)+0]>prev_nMaps){
				fprintf(stderr,"Wrong size on line %d in connection table on layer %d!",i,idxLayer);
				exit(1);
			}
			for(int j=0;j<layer->C[i*(prev_nMaps+1)+0];j++){
				fscanf(fin,"%d",&layer->C[i*(prev_nMaps+1)+j+1]);
				if(layer->C[i*(prev_nMaps+1)+j+1]<0 || layer->C[i*(prev_nMaps+1)+j+1]>=prev_nMaps){
					fprintf(stderr,"Wrong index (%d) on (%d, %d) in connection table on layer %d!",layer->C[i*(prev_nMaps+1)+j+1],i,j,idxLayer);
					exit(1);
				}
				if(j>0 && layer->C[i*(prev_nMaps+1)+j+1]<=layer->C[i*(prev_nMaps+1)+j]){
					fprintf(stderr,"Non-ascending index on (%d, %d) in connection table on layer %d!",i,j,idxLayer);
					exit(1);
				}
			}
		}
	}
	else {
		if (strcmp(s,"full")==0){
			for(int i=0;i<layer->nMaps;i++){
				layer->C[i*(prev_nMaps+1)+0]=prev_nMaps;
				for(int j=0;j<prev_nMaps;j++)
					layer->C[i*(prev_nMaps+1)+j+1]=j;
			}
			layer->C_type=full;
		}
		else if (strcmp(s,"random")==0){
			set <int> line;
			int conn,x;
			for(int i=0;i<layer->nMaps;i++){
				conn = 1 + rand() % prev_nMaps;
				layer->C[i*(prev_nMaps+1)+0]=conn;
				line.clear();
				for(int j=0;j<conn;j++){
					do{ x = rand()%prev_nMaps; }while(line.find(x)!=line.end());
					line.insert(x);
				}
				set<int>::iterator it;
				for ( x=0, it=line.begin() ; it != line.end(); x++, it++ )	layer->C[i*(prev_nMaps+1)+x+1]=*it;
			}
			layer->C_type=not_full;
		}
		else {
			fprintf(stderr,"connection table type: %s is neither full nor not_full, check the txt input file", s);
			exit(1);
		}
	}
	// allocating and computing BP_C
	layer->BP_C	= (int*)calloc(prev_nMaps * (layer->nMaps+1) , sizeof(int));
	int *C				= layer->C;		//do not move them !
	int *BP_C		= layer->BP_C;

	for(int i=0; i<layer->nMaps; i++)
		for(int j=1; j<=C[i*(prev_nMaps+1) + 0]; j++){
			BP_C[ C[i*(prev_nMaps+1) + j] * (layer->nMaps+1) + BP_C[ C[i*(prev_nMaps+1) + j] * (layer->nMaps+1) + 0] + 1] = i;
			BP_C[ C[i*(prev_nMaps+1) + j] * (layer->nMaps+1) + 0]++;
		}


		fscanf(fin,"\nActivation function: %s\n\n",s);
		if(strcmp(s,"scaled_tanh")==0)
			layer->activation=scaled_tanh;
		else
			if(strcmp(s,"soft_max")==0)
				layer->activation=soft_max;
			else;//error

			//create W
			layer->nWeights=0;
			for(int i=0;i<layer->nMaps;i++)
				layer->nWeights += 1 + layer->Kx * layer->Ky * layer->C[i*(prev_nMaps+1)+0];

			layer->W=(double*)malloc(layer->nWeights * sizeof(double));
			if(layer->W==NULL){fprintf(stderr,"Error @createConvolutionalLayer() : memory allocation (W) for layer %d.\n",idxLayer);return;}

			//weight initialization
			string b,c; 
			double amplitude=0.05;
#define UNIFORM_PLUS_MINUS_ONE ( (double)(2.0 * rand())/RAND_MAX - 1.0 )
				for (int i=0; i < layer->nWeights; i++) layer->W[i] = amplitude*UNIFORM_PLUS_MINUS_ONE;
#undef UNIFORM_PLUS_MINUS_ONE

			//allocating space for states
			layer->nNeurons = layer->nMaps * layer->Mx * layer->My;
			layer->y=(double*)malloc(layer->nNeurons * sizeof(double));
			if(layer->y==NULL){fprintf(stderr,"Error @createConvolutionalLayer() : memory allocation (y) for layer %d.\n",idxLayer);return;}
			//allocating space for deltas
			layer->d=(double*)malloc(layer->nNeurons * sizeof(double));
			if(layer->d==NULL){fprintf(stderr,"Error @createConvolutionalLayer() : memory allocation (d) for layer %d.\n",idxLayer);return;}
}

void 
NeuralNetwork::createPoolingLayer(int idxLayer, FILE *fin){
	Layer *layer		= &listLayers[idxLayer];
	Layer *prev_layer	= &listLayers[idxLayer-1];
	int prev_nMaps		= prev_layer->nMaps;

	layer->nMaps = prev_layer->nMaps;
	if(fscanf(fin,"Kernel size: %d %d\n",&layer->Kx,&layer->Ky)!=2){
		printf("Error reading Kernel sizes for layer %d.\n\7",idxLayer); exit(1);
	}
	if(fscanf(fin,"Skip: %d %d\n",&layer->Sx,&layer->Sy)!=2){
		printf("Error reading skip values for layer %d.\n\7",idxLayer); exit(1);
	}
	if(layer->Sx<=0 || layer->Sx>layer->Kx || layer->Sy<=0 || layer->Sy>layer->Ky){
		printf("Sx and Kx out of bound in layer %d.\n\7",idxLayer); exit(1);
	}
	//checking
	if(prev_layer->type!=one_kernel_convolutional && prev_layer->type!=convolutional){
		fprintf(stderr,"Max/avg pooling layer should follow a convolutional or one_kernel_convolutional layer! Please check layer %d type!",idxLayer-1);
		exit(1);
	}
	if( (prev_layer->Mx - layer->Kx) % layer->Sx != 0 || (prev_layer->My - layer->Ky) % layer->Sy != 0 ){
		fprintf(stderr,"Max/avg pooling kernels are out of bound on layer %d! Check kernel and input sizes!", idxLayer);
		exit(1);
	}
	//computing the current map sizes
	layer->Mx = (prev_layer->Mx - layer->Kx) / layer->Sx + 1;
	layer->My = (prev_layer->My - layer->Ky) / layer->Sy + 1;

	//allocating space for states
	layer->nNeurons = layer->nMaps * layer->Mx * layer->My;
	layer->y=(double*)malloc(layer->nNeurons * sizeof(double));
	if(layer->y==NULL){fprintf(stderr,"Error @createPoolingLayer() : memory allocation (y) for layer %d.\n",idxLayer);return;}
	//allocating space for deltas
	layer->d=(double*)malloc(layer->nNeurons * sizeof(double));
	if(layer->d==NULL){fprintf(stderr,"Error @createPoolingLayer() : memory allocation (d) for layer %d.\n",idxLayer);return;}
	
	if(layer->type==max_pooling){
		//allocating space for max pooling position info
		layer->max_pos=(int*)malloc(layer->nNeurons * sizeof(int));
		if(layer->max_pos==NULL){fprintf(stderr,"Error @createPoolingLayer() : memory allocation (max_pos) for layer %d.\n",idxLayer);return;}
	}
}

void
NeuralNetwork::setEta(double EtaStart, double EtaStop, double EtaFactor){
	etaStart=EtaStart;
	etaStop=EtaStop;
	etaFactor=EtaFactor;
}

void 
NeuralNetwork::setInput(unsigned char *image, unsigned char width, unsigned char height)
{
	if (width*height != listLayers[0].nNeurons){
		printf("Error: The image does not fit the neural network input!\n");
		return;
	}
	double *y=listLayers[0].y;
	for (unsigned int i=0; i<height; i++)
		for (unsigned int j=0; j<width; j++)
			y[i*width+j] = image[i*width+j] / 127.5 - 1.0;	
}

void 
NeuralNetwork::FPLayerCPU(int index){
	switch(listLayers[index].type){
		case fully_connected:	FPFullyConnectedLayerCPU(index);	break;
		case convolutional:		FPConvolutionalLayerCPU(index);		break;
		case max_pooling:		FPMaxPoolingLayerCPU(index);		break;
		default:				fprintf(stderr,"Unknown layer type: (%d) in file %s, function %s, line %d!\n\7",listLayers[index].type,__FILE__,__FUNCTION__,__LINE__); exit(1);	break; //error
	}
}

void 
NeuralNetwork::FPFullyConnectedLayerCPU(int index){
	int N					= listLayers[index].nNeurons;
	double *y				= listLayers[index].y;
	double *yprev			= listLayers[index-1].y;
	double *W				= listLayers[index].W;	
	double sum;
	int ncon				= listLayers[index-1].nNeurons+1;
	int idx;

	//if(enablelog)fprintf(flog,"OUTPUTS of layer %d:\n",index);
	switch(listLayers[index].activation){
		case scaled_tanh: 
			idx=0;
			for (int i=0; i<N; i++){
				sum = W[idx+0] * 1;	//bias
				for (int c=1; c<ncon; c++)
					sum += yprev[c-1] * W[idx+c];
				y[i] = SCALEDTANH(sum);
				idx+=ncon;
				//if(enablelog)fprintf(flog,"output[%d] = %.10f\n",in,y[in]);
			}
			break;

		case soft_max:{
			double *a = (double*)malloc(N*sizeof(double));
			for (int i=0; i<N; i++){
				sum = W[i*ncon+0] * 1;	//bias
				for (int c=1; c<ncon; c++)
					sum += yprev[c-1] * W[i*ncon+c];
				a[i] = sum;
			}
			//to check for overflow!
			double denominator=0;
			int cut=700; //DBL_MAX
			for(int i=0;i<N;i++){
				if(a[i]>cut)
					a[i]=cut;
				else 
					if(a[i]<-cut)
						a[i]=-cut;
				denominator+=exp(a[i]);
			}
			for(int i=0;i<N;i++)
				y[i] = exp(a[i])/denominator;
			free(a);
					  }break;

		default: break; //error
	}
}

int fcmpd(const void * a, const void * b){
   double va,vb;
   va=((double*)a)[1];
   vb=((double*)b)[1];
   if (fabs(va-vb)<=1e-6)
       return 0;
   return va>vb?1:-1;
} 

int fcmpResult(const void * a, const void * b){
   double va,vb;
   va=((Result*)a)->val;
   vb=((Result*)b)->val;
   if (fabs(va-vb)<=1e-6)
       return 0;
   return va<vb?1:-1;
} 

void 
NeuralNetwork::FPConvolutionalLayerCPU(int index){
	Layer *layer	=	&listLayers[index];
	int M			=	layer->nMaps;
	int Mx			=	layer->Mx;
	int My			=	layer->My;
	int Kx			=	layer->Kx;
	int Ky			=	layer->Ky;
	int Sx			=	layer->Sx;
	int Sy			=	layer->Sy;
	int *C			=	layer->C;
	double *W		=	layer->W;	
	double *y		=	layer->y;

	Layer *prevlayer=	&listLayers[index-1];
	double *yprev	=	prevlayer->y;
	int Mprev		=	prevlayer->nMaps;
	int Mxprev		=	prevlayer->Mx;
	int Myprev		=	prevlayer->My;

	double sum;
	int nconpermap	=	Kx*Ky;
	int NperMap		=	Mx*My;
	int NperMapprev	=	Mxprev*Myprev;
	int mapprev;
	double *Wmap	=	W;

	switch(listLayers[index].activation){
		case scaled_tanh: 
			for (int m=0; m<M; m++){				//for each map m
				mapprev	= C[m*(Mprev+1)+0];			//connection table
				for (int ni=0; ni<My; ni++)			//for each neuron (ni,nj) of map m
					for (int nj=0; nj<Mx; nj++){
						sum = Wmap[0] * 1;	//bias
						for(int mprev=0;mprev<mapprev;mprev++){	//for each map in the previous layer
							int idxStartN	=	C[m*(Mprev+1)+mprev+1]*NperMapprev +	ni*(Sy+1)*Mxprev + nj*(Sx+1);
							//int idxW		=	1 + C[m*(Mprev+1)+mprev+1]*nconpermap;
							int idxW		=	1 + mprev*nconpermap;
							for (int i=0; i<Ky; i++){
								for (int j=0; j<Kx; j++){
									//sum += yprev[idxStartN + i*Mxprev + j] *	Wmap[idxW + i*Kx+j];
									sum += yprev[idxStartN] *	Wmap[idxW];
									idxStartN++;
									idxW ++;
								}
								idxStartN+=Mxprev-Kx;
							}
						}
						y[m*NperMap+ni*Mx+nj] = SCALEDTANH(sum);
					}
					Wmap	+=mapprev * nconpermap + 1;
			}
			break;

		default: break; //error
	}
}

void 
NeuralNetwork::FPMaxPoolingLayerCPU(int index){
	Layer *layer	=	&listLayers[index];
	int M			=	layer->nMaps;
	int Mx			=	layer->Mx;
	int My			=	layer->My;
	int Kx			=	layer->Kx;
	int Ky			=	layer->Ky;
	double *y		=	layer->y;

	Layer *prevlayer=	&listLayers[index-1];
	double *yprev	=	prevlayer->y;
	int Mxprev		=	prevlayer->Mx;
	int Myprev		=	prevlayer->My;

	double max;
	int idx_max,idx;

	int nconpermap	=	Kx*Ky;
	int NperMap		=	Mx*My;
	int NperMapprev	=	Mxprev*Myprev;

	for (int m=0; m<M; m++){				//for each map m
		for (int ni=0; ni<My; ni++)			//for each neuron (ni,nj) of map m
			for (int nj=0; nj<Mx; nj++){
				max = -1E10;
				for (int i=0; i<Ky; i++)
					for (int j=0; j<Kx; j++){
						idx = m*NperMapprev + ni*Ky*Mxprev + nj*Kx + i*Mxprev + j;
						if(max<yprev[idx]){
							max			= yprev[idx];
							idx_max	= idx;
						}
					}
				idx							= m*NperMap+ni*Mx+nj;
				y[idx]						= max;
				layer->max_pos[idx]	= idx_max;
			}
	}
}

int 
NeuralNetwork::getClass(double *v){
	int cls = 0;
	int N = listLayers[nLayers-1].nNeurons; 
	for (int i=1; i<N; i++)
		if (v[i]>v[cls]) cls = i;
	return cls;
}

double 
NeuralNetwork::determineErrorRateCPU(IDXHandler idxh, Error *errorvector)
{
	double *image = NULL;
	int digit,label;

	int nr,count=0,outsize=listLayers[nLayers-1].nNeurons,errcount=0;
	nr = idxh.nImages;

	ERR=0.0;

	for (int k=0; k<nr; k++){
		double currentERR=0.0f;
		image = idxh.getImageAsDoubles(k);
		label = idxh.getLabel(k);
		memcpy(listLayers[0].y, image, listLayers[0].nNeurons*sizeof(double));	//copy the image to NN input
		for (int i=1; i<nLayers; i++) FPLayerCPU(i);	// forward propagate 
		double *y = listLayers[nLayers-1].y;
		digit=0;
		for (int i=1; i<outsize; i++){
			if (y[i] > y[digit])digit = i;
			switch(listLayers[nLayers-1].activation){
				case soft_max:	if(y[i]<DBL_MIN) y[i] = DBL_MIN;	currentERR -=  (label==i) * log(y[i]); break;
				case scaled_tanh:	currentERR += 0.5f * (y[i] - ((label==i)?1:-1)) * (y[i] - ((label==i)?1:-1)) ; break;
			}
		}
		ERR+=currentERR;		
		if (digit == label) count++;
		else{
			for(int j=0;j<outsize;j++)errorvector[errcount].scores[j]=y[j];
			errorvector[errcount].idx=k;
			errcount++;
		}
	}
	printf("ERR=%.6lf\n",ERR);
	return 1.0 - ((double)count)/nr;
}

double 
NeuralNetwork::determineErrorRateCPU_BIG(IDXHandler idxh, Error *errorvector){
	double *image = NULL;
	int digit,label;

	int count=0,
		outsize=listLayers[nLayers-1].nNeurons,
		errcount=0,
		nrImages = idxh.nImages;
	Result *out=(Result*)malloc(outsize*sizeof(Result));
	if(out==NULL){printf("Cannot allocate out[] in %s!\n\7",__FUNCTION__);exit(1);}

	ERR=0.0;

	for (int k=0; k<nrImages; k++){
		printf("\r%.1f%%",k*100.0/nrImages);
		double currentERR=0.0f;
		image = idxh.getImageAsDoubles(k);
		label = idxh.getLabel(k);
		memcpy(listLayers[0].y, image, listLayers[0].nNeurons*sizeof(double));	//copy the image to NN input
		for (int i=1; i<nLayers; i++) FPLayerCPU(i);	// forward propagate 
		free(image); image=NULL;
		double *y = listLayers[nLayers-1].y;
		digit=0;
		for (int i=1; i<outsize; i++){
			if (y[i] > y[digit])digit = i;
			switch(listLayers[nLayers-1].activation){
				case soft_max:	if(y[i]<DBL_MIN) y[i] = DBL_MIN;	currentERR -=  (label==i) * log(y[i]); break;
				case scaled_tanh:	currentERR += 0.5f * (y[i] - ((label==i)?1:-1)) * (y[i] - ((label==i)?1:-1)) ; break;
			}
		}
		ERR+=currentERR;		

		if (digit == label) count++;
		else{
			for(int j=0;j<outsize;j++){out[j].cls=j; out[j].val=y[j];}
			qsort(out,outsize,sizeof(Result),fcmpResult);
			for(int j=0;j<10;j++){
				errorvector[errcount].outputs[j].cls=out[j].cls;
				errorvector[errcount].outputs[j].val=out[j].val;
			}
			errorvector[errcount].idx=k;
			errcount++;
		}
	}
	printf("\rERR=%.6lf\n",ERR);
	free(out);
	return 1.0 - ((double)count)/nrImages;
}

double *
NeuralNetwork::getDesiredOutput(unsigned char label)
{
	if (listLayers == NULL) {return NULL;}

	int N = listLayers[nLayers-1].nNeurons;// the number of neurons of the last layer
	double *v = (double *)malloc(N*sizeof(double));

	for (int i=0; i<N; i++)	
		v[i] = listLayers[nLayers-1].activation==scaled_tanh ? -1.0 : 0.0;	
	v[label] = 1.0;

	return v;
}

void 
NeuralNetwork::BPLayerCPU(int idxlay){
	switch(listLayers[idxlay+1].type){
		case fully_connected:	BPFullyConnectedLayerCPU(idxlay);	break;
		case convolutional:		BPConvolutionalLayerCPU(idxlay);	break;
		case max_pooling:		BPMaxPoolingLayerCPU(idxlay);		break;
		default:				fprintf(stderr,"Unknown layer type: (%d) in file %s, function %s, line %d!\n\7",listLayers[idxlay].type,__FILE__,__FUNCTION__,__LINE__); exit(1);	break; //error
	}
}

void 
NeuralNetwork::BPFullyConnectedLayerCPU(int idxlay){
	int N							= listLayers[idxlay].nNeurons;
	double *d				= listLayers[idxlay].d;
	double *y					= listLayers[idxlay].y;
	int nextN					= listLayers[idxlay+1].nNeurons;
	double *nextDelta	= listLayers[idxlay+1].d;
	double *nextW		= listLayers[idxlay+1].W;	//next layer weights
	int ncon					= N+1;
	double eta_nextDelta;
	int i_ncon_1;
	int idx;

	if(idxlay>0)	//BP deltas, except for the input layer
		for (int j=0; j<N; j++){
			d[j] = 0.0;
			idx=j+1;
			for (int i=0; i<nextN; i++){
				//d[j] += nextW[i*ncon+j+1] * nextDelta[i];
				d[j] += nextW[idx] * nextDelta[i];
				idx+=ncon;
			}
			d[j] *= DSCALEDTANH(y[j]);
		}
		//updating weights
		for (int i=0; i<nextN; i++){
			eta_nextDelta = eta * nextDelta[i]; i_ncon_1 = i*ncon+1;	//precomputing indexes
			nextW[i*ncon+0] -= eta_nextDelta;	//bias
			for (int j=0; j<N; j++)
				nextW[i_ncon_1+j] -= eta_nextDelta*y[j];
		}
}

void 
NeuralNetwork::BPConvolutionalLayerCPU(int idxlayer){
	Layer *layer	=	&listLayers[idxlayer+1];
	int N			=	layer->nNeurons;
	int M			=	layer->nMaps;
	int Mx			=	layer->Mx;
	int My			=	layer->My;
	int Kx			=	layer->Kx;
	int Ky			=	layer->Ky;
	int Sx			=	layer->Sx;
	int Sy			=	layer->Sy;
	int *C			=	layer->C;
	double *W		=	layer->W;	
	double *y		=	layer->y;
	double *d		=	layer->d;

	Layer *prevlayer=	&listLayers[idxlayer];
	double *yprev	=	prevlayer->y;
	double *dprev	=	prevlayer->d;
	int Mprev		=	prevlayer->nMaps;
	int Mxprev		=	prevlayer->Mx;
	int Myprev		=	prevlayer->My;
	int Nprev		=	prevlayer->nNeurons;

	double delta,eta_delta;
	int nconpermap	=	Kx*Ky;
	int NperMap		=	Mx*My;
	int NperMapprev	=	Mxprev*Myprev;
	int mapprev;
	double *Wmap	=	W;

	if(idxlayer>0){	//BP deltas, except for the input layer
		memset((void*)dprev,0,Nprev*sizeof(double));
		for (int m=0; m<M; m++){				//for each map m
			mapprev	= C[m*(Mprev+1)+0];			//connection table (# of connected maps)
			for (int ni=0; ni<My; ni++)			//for each neuron (ni,nj) of map m
				for (int nj=0; nj<Mx; nj++){
					delta = d[m*NperMap + ni*Mx + nj];
					for(int mprev=0; mprev<mapprev; mprev++){	//for each map in the previous layer
						int idxStartN	=	C[m*(Mprev+1)+mprev+1]*NperMapprev;	//indexes of neurons in map mprev starts at this index
						int idxInMap	=	ni*(Sy+1)*Mxprev + nj*(Sx+1);
						idxStartN += idxInMap;
						//int idxW		=	1 + C[m*(Mprev+1)+mprev+1]*nconpermap;
						int idxW		=	1 + mprev*nconpermap;
						for (int i=0; i<Ky; i++){
							for (int j=0; j<Kx; j++){
								//dprev[idxStartN + idxInMap + i*Mxprev + j] += delta * Wmap[idxW + i*Kx+j];
								dprev[idxStartN++] += delta * Wmap[idxW++];
								//idxStartN++;
								//idxW++;
							}
							idxStartN+=Mxprev-Kx;
						}
					}
				}
				Wmap += mapprev * nconpermap + 1;
		}
		for(int i=0;i<Nprev;i++)dprev[i] *= DSCALEDTANH(yprev[i]);
	}

	//updating weights
	Wmap = W;
	for (int m=0; m<M; m++){				//for each map m
		mapprev	= C[m*(Mprev+1)+0];			//connection table (# of connected maps)
		for (int ni=0; ni<My; ni++)			//for each neuron (ni,nj) of map m
			for (int nj=0; nj<Mx; nj++){
				eta_delta = eta * d[m*NperMap + ni*Mx + nj];
				Wmap[0] -= eta_delta * 1;	//for bias
				for(int mprev=0; mprev<mapprev; mprev++){	//for each map in the previous layer
					int idxStartN	=	C[m*(Mprev+1)+mprev+1]*NperMapprev;	//indexes of neurons in map mprev starts at this index
					int idxInMap	=	ni*(Sy+1)*Mxprev + nj*(Sx+1);
					idxStartN	+= idxInMap;
					//int idxW		=	1 + C[m*(Mprev+1)+mprev+1]*nconpermap;
					int idxW		=	1 + mprev*nconpermap;
					/*	for (int i=0; i<Ky; i++)
					for (int j=0; j<Kx; j++)
					Wmap[idxW + i*Kx+j] -= eta_delta * yprev[idxStartN + idxInMap + i*Mxprev + j];*/
					for (int i=0; i<Ky; i++){
						for (int j=0; j<Kx; j++){
							//dprev[idxStartN + idxInMap + i*Mxprev + j] += delta * Wmap[idxW + i*Kx+j];
							Wmap[idxW++] -= eta_delta * yprev[idxStartN++];
							//idxStartN++;
							//idxW++;
						}
						idxStartN+=Mxprev-Kx;
					}
				}
			}
			Wmap += mapprev * nconpermap + 1;
	}
}

void 
NeuralNetwork::BPMaxPoolingLayerCPU(int idxlayer){
	Layer *layer		=	&listLayers[idxlayer+1];
	int N					=	layer->nNeurons;
	double *d			=	layer->d;
	Layer *prevlayer=	&listLayers[idxlayer];
	double *dprev	=	prevlayer->d;
	int Nprev			=	prevlayer->nNeurons;
	int *max_pos	=	layer->max_pos;
	
	if(idxlayer>0){	//BP deltas, except for the input layer
		memset((void*)dprev,0,Nprev*sizeof(double));
		for (int n=0; n<N; n++)
			dprev[max_pos[n]] = d[n];
	}
}

void 
NeuralNetwork::correctWeightsCPU(double *image, unsigned char label){
	if(listLayers[0].IP==0)
		listLayers[0].y = image;
	else
		memcpy(listLayers[0].y, image, listLayers[0].nNeurons*sizeof(double));	//copy the image to NN input
	for (int i=1; i<nLayers; i++) FPLayerCPU(i);				// forward propagate 
	double *y = listLayers[nLayers-1].y;
	double *t = getDesiredOutput(label);
	if (y == NULL || t == NULL){
		printf("Error @correctWeightsCPU: memory allocation (t or u) !\n");
		return;
	}
	if (getClass(y) == label) count++;

	double *currDelta	= listLayers[nLayers-1].d;
	int N				= listLayers[nLayers-1].nNeurons;

	// compute diferentials for the output layer 

	double currentERR = 0.0;
	for (int k=0; k<N; k++){
		switch(listLayers[nLayers-1].activation){
			case soft_max:
				if(y[k]<FLT_MIN) y[k] = FLT_MIN;
				currentERR -=  (label==k) * log(y[k]);
				currDelta[k] = y[k] - t[k];
				break;
			case scaled_tanh:
				currentERR += 0.5f * (y[k] - t[k]) * (y[k] - t[k]);
				currDelta[k] = DSCALEDTANH(y[k]) * (y[k] - t[k]);
				break;
			default: break;
		}
	}
	ERR += currentERR;

	for (int i=nLayers-2; i>=0; i--) BPLayerCPU(i);
	free(t);
}

void reverse(unsigned  &sig){
	sig=(sig<<24) | ((sig<<8)&0x00FF0000) | ((sig>>8)&0x0000FF00) | (sig>>24);
}

void
NeuralNetwork::TrainOneEpochCPU(IDXHandler idxh){
	unsigned  width  = idxh.width;
	unsigned  height = idxh.height;
	double*	  image	 = NULL;
	unsigned char label;
	unsigned  nrImages,i;
	
	nrImages = idxh.nImages;
	ERR=0;
	/*clock_t t1,t2;
	t1 = clock(); */
	for (i=0; i<nrImages; i++){
		if(nrImages>=10 && i%(nrImages/1000)==0)
			printf("\r%.2f%%",i*100.0/nrImages);
		if(i%idxh.bufferSize==0){
			idxh.collectData(i);
		}
		//printf("%d ",i);
		image = idxh.getImageAsDoubles(i); 

		//dump images
		if(epoch==0 && i==0){
			int bufferSize=listLayers[0].nMaps;
			int size=listLayers[0].Mx*listLayers[0].My*bufferSize;
			char filename[1000], timebuf[26];
			time_t ltime;
			time( &ltime );
			ctime_s(timebuf, 26, &ltime);
			sprintf(filename,"img_cpu%d-%ld.idx",i,ltime);

			FILE *f; if ((f=fopen(filename,"wb")) == NULL){printf("Error: can not open file %s\n",filename);return;}
			unsigned  signature = 0x8030000; //...803 caci am nevoie sa salvez doar setul de antrenare
			reverse(signature);    fwrite(&signature,sizeof(signature),1,f);
			unsigned  r_nImages = bufferSize;
			unsigned  r_height  = listLayers[0].My;
			unsigned  r_width   = listLayers[0].Mx;
			reverse(r_nImages); fwrite(&r_nImages,sizeof(r_nImages),1,f);
			reverse(r_height);  fwrite(&r_height,sizeof(r_height),1,f);
			reverse(r_width);   fwrite(&r_width,sizeof(r_width),1,f);

			unsigned char *tempbuffer = (unsigned char *)malloc(size*sizeof(unsigned char));
			if (tempbuffer == NULL) {printf("Error: malloc tempbuffer IDX in function %s, file %s, line %d.\n",__FUNCTION__,__FILE__,__LINE__);return;}

			int row,col;
			float f_aux;
			int i_aux;
			for (int k=0; k<bufferSize; k++){
				for (row=0; row<listLayers[0].My; row++)
					for (col=0; col<listLayers[0].Mx; col++){
						f_aux = 127.5f + 127.5f*image[k*listLayers[0].Mx*listLayers[0].My + row*listLayers[0].Mx + col];
						i_aux = (int)f_aux;
						tempbuffer[k*listLayers[0].Mx*listLayers[0].My + row*listLayers[0].Mx + col] = (unsigned char) i_aux;
					}
			}
			fwrite(tempbuffer,sizeof(unsigned char),size,f);
			fclose(f);
			free(tempbuffer);
		}
        //end dump images
		
		label = idxh.getLabel(i);	
		
		correctWeightsCPU(image,label);
		free(image);
		
	}
}

void 
NeuralNetwork::TrainCPU(IDXHandler idxh_train, IDXHandler idxh_validation, IDXHandler idxh_test)
{
	clock_t t1,t2;
	double elapsedTime;
	unsigned  skipped=0;
	int last_epoch=0;
	Error *errorvectorV=(Error*)malloc((idxh_train.nImages)*sizeof(Error));
	Error *errorvectorT=(Error*)malloc((idxh_test.nImages)*sizeof(Error));

	// set initial eta to etaStart for newly created net
	eta=etaStart;

	epoch=0;
	while(eta>etaStop){
		double test_err,validation_err;
		count=0;

		printf("\nEpoch: %d\teta=%f\n",epoch,eta);
	
		idxh_train.scramble();
		
		t1=clock();
		TrainOneEpochCPU(idxh_train);
		t2=clock();
		elapsedTime = (double)(t2-t1)/CLOCKS_PER_SEC;
		printf("ERROR = %.7lf\n",ERR);
		printf("Error rate on training set:\t%.2f%%\t",100.0*(idxh_train.nImages-count)/idxh_train.nImages);
		printf("Elapsed Time = %5.3lf\n",elapsedTime);

		t1=clock();validation_err = determineErrorRateCPU_BIG(idxh_validation,errorvectorV);t2=clock();
		printf("Error Rate on validation set: \t%.3f%%\t",100*validation_err);printf("Elapsed Time = %5.3lf\n",(double)(t2-t1)/CLOCKS_PER_SEC);
		t1=clock();test_err = determineErrorRateCPU_BIG(idxh_test,errorvectorT);t2=clock();
		printf("Error Rate on test set: \t%.3f%%\t",100*test_err);printf("Elapsed Time = %5.3lf\n",(double)(t2-t1)/CLOCKS_PER_SEC);

		int write_test=0, write_validation=0;
		if(epoch==0){
			best_validation = validation_err;
			best_test = test_err;
			write_test = write_validation = 1;
		}
		else{
			if(best_test > test_err){
				best_test = test_err;
				write_test = 1;
			}
			if(best_validation > validation_err){
				best_validation = validation_err;
				write_validation = 1;
			}
			if(best_test > test_err && best_validation > validation_err)printf("\7");
		}

		if(write_test || write_validation)
			updateBestResults(write_test,write_validation,errorvectorV,errorvectorT,(float)validation_err,(float)test_err,idxh_validation,idxh_test,false);

		eta = eta*etaFactor;
		epoch++;
	}
	free(errorvectorV);
	free(errorvectorT);
}

int NeuralNetwork::getOutputSize(void){
	return listLayers[nLayers-1].nNeurons;
}

int NeuralNetwork::getInputSize(void){
	return listLayers[0].nNeurons;
}

void NeuralNetwork::delete_files(char *base, char *type){
	char temp[1000];

	sprintf(temp,"%s-%s-%s.idx",NNname,type,base);remove(temp);
	sprintf(temp,"%s-%s-%s.txt",NNname,type,base);remove(temp);
}

void NeuralNetwork::write_error_files(char *type, char *base, Error *errorvector, int errors, IDXHandler idxh){
    char name[1000];
	int CR[10]={0};		//compute different errors
    sprintf(name,"%s-%s-%s.idx",NNname,type,base);
	idxh.saveErrorImages(name,errorvector,errors);
    sprintf(name,"%s-%s-%s.txt",NNname,type,base);
    FILE *ferr=fopen(name,"wt");
    // confMat [nr_classes,nr_classes] row=true label, column=prediction
	int *confMat=(int*)calloc(idxh.n_labels*idxh.n_labels,sizeof(int));
    
    if (confMat==NULL){
        printf("Cannot allocate confMat.\n\7");
        exit(1);
    }
	int nscores = 10;
    for(int i=0;i<errors;i++){
        fprintf(ferr,"\nindex=%d\tlabel=%d\n", errorvector[i].idx, idxh.label_map[idxh.getLabel(errorvector[i].idx)]);
        for(int j=0;j<nscores;j++){
			fprintf(ferr,"%d - %f\n", idxh.label_map[errorvector[i].outputs[j].cls], errorvector[i].outputs[j].val);
     		if(errorvector[i].outputs[j].cls==idxh.getLabel(errorvector[i].idx))
				for(int k=j;k<10;k++)CR[k]++;
        }
        // now fill the confusion matrix
		confMat[idxh.getLabel(errorvector[i].idx)*idxh.n_labels+errorvector[i].outputs[0].cls]++;
    }
    // write confusion table
    fprintf(ferr,"\nConfusion matrix [%d,%d], total number of errors: %d\n",idxh.n_labels,idxh.n_labels,errors);

	double rec_rate=0;
    for (int i=0;i<idxh.n_labels;i++){
        int errors=0;
        for (int j=0;j<idxh.n_labels;j++){
			errors+=confMat[i*idxh.n_labels+j];
			fprintf(ferr,"%d\t",confMat[i*idxh.n_labels+j]);
        }
        fprintf(ferr,"\t%d (%.2f%% out of %d)\n",errors,100.0*errors/idxh.samplesPerClass[i],idxh.samplesPerClass[i]);
		rec_rate+=((double)idxh.samplesPerClass[i]-errors)/idxh.samplesPerClass[i];
    }
	printf("Average per class recognition rate: %.2f%%\n",100.0*rec_rate/idxh.n_labels);
	fprintf(ferr,"\nAverage per class recognition rate: %.2f%%\n",100.0*rec_rate/idxh.n_labels);
	fprintf(ferr,"\n");
	for(int i=0;i<__min(nscores,10);i++){
		fprintf(ferr,"CR%d: %.3f%%\n",i+1,100.0*(idxh.nImages-errors+CR[i])/idxh.nImages);
	}
	free(confMat);
    fclose(ferr);
} 

void NeuralNetwork::updateBestResults(int write_test, int write_validation, Error *errorVectorV, Error *errorVectorT, float validation_err, float test_err, IDXHandler &idxh_validation, IDXHandler &idxh_test, bool isGPU){
	char temp[1000];
	if(write_validation){
		
		delete_files(best_validation_name,"BVerr");
		delete_files(best_validation_name,"TfBVerr");

		sprintf(best_validation_name,"V-%.3f-T-%.3f-epoch-%d",100*validation_err,100*test_err,epoch);
		write_error_files("BVerr",		best_validation_name,	errorVectorV,	myroundf(validation_err*idxh_validation.nImages),idxh_validation);
		write_error_files("TfBVerr",	best_validation_name,	errorVectorT,	myroundf(test_err*idxh_test.nImages),			idxh_test);
	}

	if(write_test){
		
		delete_files(best_test_name,"BTerr");

		sprintf(best_test_name,"V-%.3f-T-%.3f-epoch-%d",100*validation_err,100*test_err,epoch);
		write_error_files("BTerr",		best_test_name,			errorVectorT,	myroundf(test_err*idxh_test.nImages),			idxh_test);
	}
}

int cmpErr(const void *a, const void *b){
	int class_a = (int)(((Error*)a)->scores[43]);
	int class_b = (int)(((Error*)b)->scores[43]);
	if(class_a == class_b)
		return ((Error*)a)->idx - ((Error*)b)->idx;
	else
		return  class_a - class_b;
}
